/*
 * Library interface to Linux x86 Performance-Monitoring Counters.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include "libperfctr.h"
#define CONFIG_PERFCTR
#include "linux/prctl.h"

static unsigned char events_format;
static int have_rdpmc = 0;	/* false on old Intel P5 without MMX */
static int have_rdtsc = 0;	/* false on WinChip when PMCs are used */
static int init_done = 0;

static int check_init(void)
{
    if( !init_done ) {
	struct perfctr_info info;

	if( prctl(PR_PERFCTR, PERFCTR_INFO, -1, &info) < 0 )
	    return -1;
	events_format = info.events_format;
	have_rdpmc = info.have_rdpmc;
	have_rdtsc = info.have_rdtsc;
	init_done = 1;
    }
    return 1;
}

static int sys_perfctr(unsigned cmd, int whom, void *arg)
{
    if( check_init() < 0 )
	return -1;
    return prctl(PR_PERFCTR, cmd, whom, arg, 0);
}

int perfctr_info(struct perfctr_info *info)
{
    return sys_perfctr(PERFCTR_INFO, -1, info);
}

static int self_fd = -1;
static volatile struct perfctr_state kstate;	/* updated by kernel */

void perfctr_detach_self(void)
{
    if( self_fd >= 0 ) {
	close(self_fd);
	self_fd = -1;
	sys_perfctr(PERFCTR_UNLINK, -1, NULL);
	kstate.status = -1;
    }
}

int perfctr_attach_rdonly_self(void)
{
    if( self_fd >= 0 ) {
	errno = EBUSY;
	return -1;
    }
    self_fd = sys_perfctr(PERFCTR_ATTACH_RDONLY, getpid(), NULL);
    return self_fd;
}

int perfctr_attach_rdwr_self(void)
{
    if( self_fd >= 0 ) {
	errno = EBUSY;
	return -1;
    }
    self_fd = sys_perfctr(PERFCTR_ATTACH_RDWR, getpid(), NULL);
    return self_fd;
}

static __inline__ int check_self_fd(void)
{
    return (self_fd < 0) ? perfctr_attach_rdwr_self() : 1;
}

union u64_conversion {	/* safe type-punning */
    unsigned int ui[2];
    unsigned long long ull;
};

static __inline__ unsigned long long rdtsc64(void)
{
    union u64_conversion u;
    __asm__ __volatile__("rdtsc" : "=a"(u.ui[0]), "=d"(u.ui[1]));
    return u.ull;
}

static __inline__ unsigned long long rdpmc64(unsigned int ctr)
{
    union u64_conversion u;
    __asm__ __volatile__("rdpmc" : "=a"(u.ui[0]), "=d"(u.ui[1]) : "c"(ctr));
    return u.ull;
}

int perfctr_read_self(struct perfctr_state *state)
{
    if( check_self_fd() < 0 )
	return -1;
    if( have_rdpmc && kstate.status > 0 ) {
	unsigned long long tsc_start, tsc, pmc[2];
	do {
	    tsc_start = kstate.ctrs.tsc_start;
	    if( have_rdtsc )	/* WinChip lossage */
		tsc = rdtsc64();
	    pmc[0] = rdpmc64(0);
	    pmc[1] = rdpmc64(1);
	    if( have_rdtsc )	/* WinChip lossage */
		state->ctrs.tsc = kstate.ctrs.tsc + (tsc - tsc_start);
	    state->ctrs.pmc[0] = kstate.ctrs.pmc[0] + pmc[0];
	    state->ctrs.pmc[1] = kstate.ctrs.pmc[1] + pmc[1];
	} while( tsc_start != kstate.ctrs.tsc_start );
	state->ctrs.tsc_start = tsc_start;
	state->control = kstate.control;
	state->status = kstate.status;
	return 0;
    } else
	return sys_perfctr(PERFCTR_READ, self_fd, state);
}

int perfctr_control_self(struct perfctr_control *control)
{
    if( check_self_fd() < 0 )
	return -1;
    control->user = (struct perfctr_state*)&kstate;
    if( sys_perfctr(PERFCTR_CONTROL, self_fd, control) < 0 )
	return -1;
    return 0;
}

int perfctr_stop_self(void)
{
    int ret;

    if( check_self_fd() < 0 )
	return -1;
    ret = sys_perfctr(PERFCTR_STOP, self_fd, NULL);
    kstate.status = 0;
    return ret;
}

int perfctr_unlink_self(void)
{
    int ret = sys_perfctr(PERFCTR_UNLINK, -1, NULL);
    kstate.status = -1;
    return ret;
}

static int perfctr_attach_simple(unsigned cmd, int pid, struct perfctr_control *control)
{
    int fd = sys_perfctr(cmd, pid, NULL);
    if( fd >= 0 && control && sys_perfctr(PERFCTR_CONTROL, fd, control) < 0 ) {
	close(fd);
	fd = -1;
    }
    return fd;
}

static int perfctr_attach_hairy(unsigned cmd, int pid, struct perfctr_control *control)
{
    int fd, oerrno;

    if( pid == getpid() )
	return perfctr_attach_simple(cmd, pid, NULL);
    if( ptrace(PTRACE_ATTACH, pid, 0, 0) < 0 )
	return -1;
    if( waitpid(pid, NULL, WUNTRACED) < 0 ) {
	fd = -1;
	goto out_detach;
    }
    fd = perfctr_attach_simple(cmd, pid, control);
 out_detach:
    oerrno = errno;
    ptrace(PTRACE_DETACH, pid, 0, SIGCONT);
    errno = oerrno;
    return fd;
}

int perfctr_attach_rdonly(int pid)
{
    return perfctr_attach_hairy(PERFCTR_ATTACH_RDONLY, pid, NULL);
}

int perfctr_attach_rdwr(int pid, struct perfctr_control *control)
{
    return perfctr_attach_hairy(PERFCTR_ATTACH_RDWR, pid, control);
}

int perfctr_read_attached(int fd, struct perfctr_state *state)
{
    return sys_perfctr(PERFCTR_READ, fd, state);
}

int perfctr_control_attached(int fd, struct perfctr_control *control)
{
    return sys_perfctr(PERFCTR_CONTROL, fd, control);
}

int perfctr_stop_attached(int fd)
{
    return sys_perfctr(PERFCTR_STOP, fd, NULL);
}

int perfctr_unlink_attached(int fd)
{
    return sys_perfctr(PERFCTR_UNLINK, fd, NULL);
}

const char *perfctr_events_name(void)
{
    if( check_init() < 0 )
	return "Function not implemented";	/* ENOSYS */
    switch( events_format ) {
      case PERFCTR_EVENTS_INTEL_P5:
        return "Intel Pentium";
      case PERFCTR_EVENTS_INTEL_P5MMX:
        return "Intel Pentium MMX";
      case PERFCTR_EVENTS_INTEL_P6:
        return "Intel Pentium Pro";
      case PERFCTR_EVENTS_INTEL_PII:
        return "Intel Pentium II";
      case PERFCTR_EVENTS_INTEL_PIII:
        return "Intel Pentium III";
      case PERFCTR_EVENTS_CYRIX_MII:
        return "Cyrix MII";
      case PERFCTR_EVENTS_WINCHIP_C6:
	return "WinChip C6";
      case PERFCTR_EVENTS_WINCHIP_2:
	return "WinChip 2";
      default:
        return "?";
    }

}
unsigned int perfctr_evntsel_num_insns(void)
{
    /* This is terribly naive. Assumes only one event will be
     * selected, at CPL > 0.
     */
    if( check_init() < 0 )
	return (unsigned int)-1;
    switch( events_format ) {
      case PERFCTR_EVENTS_INTEL_P5:
      case PERFCTR_EVENTS_INTEL_P5MMX:
      case PERFCTR_EVENTS_CYRIX_MII:
        /* event 0x16, count at CPL 3 */
        return 0x16 | (2 << 6);
      case PERFCTR_EVENTS_INTEL_P6:
      case PERFCTR_EVENTS_INTEL_PII:
      case PERFCTR_EVENTS_INTEL_PIII:
        /* event 0xc0, count at CPL > 0, ENable */
        return 0xc0 | (1 << 16) | (1 << 22);
      case PERFCTR_EVENTS_WINCHIP_C6:
        return 0x02;
      case PERFCTR_EVENTS_WINCHIP_2:
        return 0x16;
      default:
        fprintf(stderr, "events_format %u not recognized\n", events_format);
        exit(1);
    }
}
